"""
"""
import urllib

from Products.CMFCore.utils import getToolByName
from zope import component
from zope import interface

from interfaces import IPaypalStandardOptions, IPaypalStandardProcessor

from Products.PloneGetPaid.interfaces import IGetPaidManagementOptions
from getpaid.core import interfaces as GetPaidInterfaces

_sites = {
    "Sandbox": "www.sandbox.paypal.com",
    "Production": "www.paypal.com",
    }

class PaypalStandardProcessor( object ):
   
    interface.implements( IPaypalStandardProcessor )

    options_interface = IPaypalStandardOptions

    def __init__( self, context ):
        self.context = context

    def cart_post_button( self, order ):
        options = IPaypalStandardOptions( self.context )
        siteroot = getToolByName(self.context, "portal_url").getPortalObject()
        manage_options = IGetPaidManagementOptions( siteroot )        
        cartitems = []
        idx = 1
        _button_form = """<form style="display:inline;" action="https://%(site)s/cgi-bin/webscr" method="post" id="paypal-button">
<input type="hidden" name="cmd" value="_cart" />
<input type="hidden" name="upload" value="1" />
<input type="hidden" name="business" value="%(merchant_id)s" />
<input type="hidden" name="currency_code" value="%(currency)s" />
<input type="hidden" name="return" value="%(return_url)s" />
<input type="hidden" name="cbt" value="Return to %(store_name)s" />
<input type="hidden" name="rm" value="2" />
<input type="hidden" name="notify_url" value="%(IPN_url)s" />
<input type="hidden" name="invoice" value="%(order_id)s" />
<input type="hidden" name="no_note" value="1" />
%(cart)s
<input type="image" src="http://%(site)s/en_US/i/btn/x-click-but01.gif"
    name="submit"
    alt="Make payments with PayPal - it's fast, free and secure!" />
</form>
"""
        _button_cart = """<input type="hidden" name="item_name_%(idx)s" value="%(item_name)s" />
<input type="hidden" name="item_number_%(idx)s" value="%(item_number)s" />
<input type="hidden" name="amount_%(idx)s" value="%(amount)s" />
<input type="hidden" name="quantity_%(idx)s" value="%(quantity)s" />
"""
        
        for item in order.shopping_cart.values():
            v = _button_cart % {"idx": idx,
                                "item_name": item.name,
                                "item_number" : item.product_code,
                                "amount": item.cost,
                                "quantity": item.quantity,}
            cartitems.append(v)
            idx += 1
        siteURL = siteroot.absolute_url()
        # having to do some magic with the URL passed to Paypal so their system replaies properly
        returnURL = "%s/@@getpaid-thank-you" % siteURL
        IPNURL = "%s/%s" % (siteURL, urllib.quote_plus("@@getpaid-paypal-ipnreactor"))
        formvals = {
            "site": _sites[options.server_url],
            "merchant_id": options.merchant_id,
            "cart": ''.join(cartitems),
            "return_url": returnURL,
            "currency": options.currency,
            "IPN_url" : IPNURL,
            "order_id" : order.order_id,
            "store_name": manage_options.store_name,
            }
        return _button_form % formvals
    
    def capture(self, order, price):
        # always returns async - just here to make the processor happy
        return GetPaidInterfaces.keys.results_async

    def authorize( self, order, payment ):
        pass


from UserDict import DictMixin

main_mapping = { 'address_city':'address_city',
                 'address_country':'address_country',
                 'address_country_code':'address_country_code',
                 'address_name':'address_name',
                 'address_state':'address_state',
                 'address_status':'address_status',
                 'address_street':'address_street',
                 'address_zip':'address_zip',
                 'first_name':'first_name',
                 'last_name':'last_name',
                 'payer_business_name':'business_name',
                 'payer_email':'email',
                 'payer_id':'payer_id',
                 'payer_status':'payer_status',
                 'contact_phone':'phone',
                 'residence_country':'residence_country',
                 'business':'my_business',
                 'receiver_email':'receiver_email',
                 'receiver_id':'receiver_id',
                 'custom':'custom',
                 'invoice':'invoice',
                 'memo':'memo',
                 'option_name_1':'option_name_1',
                 'option_name_2':'option_name_2',
                 'option_selection1':'option_selection1',
                 'option_selection2':'option_selection2',
                 'tax':'tax',
                 'auth_id':'auth_id',
                 'auth_exp':'auth_exp',
                 'auth_amount':'auth_amount',
                 'num_cart_items':'num_cart_items',
                 'parent_txn_id':'parent_txn_id',
                 'payment_date':'payment_date',
                 'payment_status':'payment_status',
                 'payment_type':'payment_type',
                 'pending_reason':'pending_reason',
                 'reason_code':'reason_code',
                 'remaining_settle':'remaining_settle',
                 'shipping_method':'shipping_method',
                 'shipping':'shipping',
                 'transaction_entity':'transaction_entity',
                 'txn_id':'txn_id',
                 'txn_type':'txn_type',
                 'exchange_rate':'exchange_rate',
                 'mc_currency':'mc_currency',
                 'mc_fee':'mc_fee',
                 'mc_gross':'mc_gross',
                 'mc_shipping#':'mc_total_shipping',
                 'mc_handling#':'mc_total_handling',
                 'payment_fee':'payment_fee',
                 'payment_gross':'payment_gross',
                 'settle_amount':'settle_amount',
                 'settle_currency':'settle_currency',
                 'test_ipn':'test_ipn',
                 }

auction_mapping = {'auction_buyer_id':'auction_buyer_id',
                   'auction_closing_date':'auction_closing_date',
                   'auction_multi_item':'auction_multi_item',
                   'for_auction':'for_auction',
                   }

subscription_mapping = {'subscr_date':'subscription_date',
                        'subscr_effective':'subscription_effective',
                        'period1':'subscription_period1',
                        'period2':'subscription_period2',
                        'period3':'subscription_period3',
                        'amount1':'subscription_amount1',
                        'amount2':'subscription_amount2',
                        'amount3':'subscription_amount3',
                        'mc_amount1':'subscription_mc_amount1',
                        'mc_amount2':'subscription_mc_amount2',
                        'mc_amount3':'subscription_mc_amount3',
                        'mc_currency':'subscription_mc_currency',
                        'recurring':'subscription_recurring',
                        'reattempt':'subscription_reattempt',
                        'retry_at':'subscription_retry_at',
                        'recur_times':'subscription_recur_times',
                        'username':'subscription_username',
                        'password':'subscription_password',
                        'subscr_id':'subscription_id',
                        }

dispute_mapping = {'case_id':'case_id',
                   'case_type':'case_type',
                   'case_creation_date':'case_creation_date',
                   'reason_code':'reason_code',
                   }

cart_item_mapping = {'item_name%s':'item_name',
                     'item_number%s':'item_number',
                     'quantity%s':'quantity',
                     'mc_gross%s':'mc_gross',
                     'mc_handling%s':'mc_handling',
                     'mc_shipping%s':'mc_shipping',
                     'option_name1%s':'option_name1',
                     'option_name2%s':'option_name2',
                     'option_selection1%s':'option_selection1',
                     'option_selection2%s':'option_selection2',
                     }

payment_txn_types = ['cart', 'express_checkout', 'merch_pmt', 'send_money', 
                     'virtual_terminal', 'web_accept']

subscription_txn_types = ['subscr-failed', 'subscr-cancel', 'subscr-payment', 
                          'subscr-signup', 'subscr-eot', 'subscr-modify']

dispute_txn_types = ['new_case', 'adjustment']

class CartItem(object):

    item_name = None
    item_number = None
    quantity = None
    mc_gross = None
    mc_handling = None
    mc_shipping = None
    option_name1 = None
    option_name2 = None
    option_selection1 = None
    option_selection2 = None

class Notification(DictMixin):
    
    shopping_cart = {}
    mass_payments = {}
    _form_variables = []

    def __init__(self, request = None):
        if request == None:
            return
        self.parse(request)
    
    def __repr__(self):
        return "<Notification at %s>" % id(self)

    # define a read-only dict interface to make it easier to work with upstream
    def __getitem__(self, item):
        return getattr(self, item)
    
    def __setitem__(self, item, value):
        return None
    
    def __delitem__(self, item):
        return None

    def keys(self):
        return self._form_variables
    # end of DictMixin classes

    def _do_parse(self, request, mapping):
        for item in mapping.keys():
            if request.has_key(item):
                self._form_variables.append(mapping[item])
                setattr(self, mapping[item], request[item])
            else:
                setattr(self, mapping[item], None)

    def parse(self, request):
        self._do_parse(request, main_mapping)
        
        # do specific parse methods for various types of IPN notifications
        if request.has_key('for_auction') and request['for_auction'] == 'true':
            self._parse_auction(request)
        
        if self.txn_type == 'cart':
            self._parse_cart(request)
        
        if self.txn_type == "masspay":
            self._parse_masspay(request)
            return
        
        if self.txn_type in subscription_txn_types:
            self._parse_subscription(request)
            return
        
        if self.txn_type in dispute_txn_types:
            self._parse_dispute(request)
            return
        
    def _parse_cart(self, request):
        self._form_variables.append('shopping_cart')
        if self.num_cart_items is not None:
            cartitems = range(1, int(self.num_cart_items)+1)
        else:
            cartitems = [1,]
        for i in cartitems:
            if request.has_key('item_number%s' % i):
                cartkey = request['item_number%s' % i]
            else:
                cartkey = "%s" % i
            self.shopping_cart[cartkey] = CartItem()
            for item in cart_item_mapping.keys():
                if request.has_key(item % i):
                    # parse out the shopping cart into each variable
                    setattr(self.shopping_cart[cartkey], cart_item_mapping[item], request[item % i])


    def _parse_auction(self, request):
        self._do_parse(request, auction_mapping)

    def _parse_subscription(self, request):
        self._do_parse(request, subscription_mapping)

    def _parse_masspay(self, request):
        """Stubbed out for now"""
    
    def _parse_dispute(self, request):
        self._do_parse(request, dispute_mapping)



import urllib, urllib2
import socket
import logging
import re

from Products.CMFCore.utils import getToolByName
from Products.Five.browser import BrowserView
from zope.component import getUtility

from getpaid.core.interfaces import IOrderManager

from getpaid.paypal.interfaces import IPaypalStandardOptions
from getpaid.paypal.paypal import _sites

from notification import Notification

logger = logging.getLogger("Plone")

class IPNListener(BrowserView):
    """Listener for Paypal IPN notifications - registered as a page view
    """

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.portal = getToolByName(self.context, 'portal_url').getPortalObject()
    
    def process(self):
        this_notification = Notification(self.request)
        is_valid_IPN = self.verify()
        if not is_valid_IPN:
            logger.info('getpaid.paypal: received bogus IPN')
            return
        order_manager = getUtility(IOrderManager)
        if this_notification.invoice in order_manager:
            order = order_manager.get(this_notification.invoice)
            if not self.compare_cart(this_notification, order):
                logger.info('getpaid.paypal: received IPN that does match order number %s' % this_notification.invoice)
                # bad IPN - do not apply to transaction
                return
            if this_notification.payment_status == 'Completed':
                self.fill_in_order_data(this_notification, order)
                order.finance_workflow.fireTransition('charge-charging')
                logger.info('getpaid.paypal: received successful IPN payment notification for order %s' % this_notification.invoice)
                return
            if this_notification.payment_status == 'Failed':
                order.finance_workflow.fireTransition('decline-charging')
                logger.info('getpaid.paypal: received unsuccessful IPN payment notification for order %s' % this_notification.invoice)
                return
            # IPN not of interest to us right now
            logger.info('getpaid.paypal: received IPN for order %s that is not of interest - txn_type "%s", payment_status "%s"' % (this_notification.invoice, this_notification.txn_type, this_notification.payment_status))
            return
        # invoice not in cart
        logger.info('getpaid.paypal: received IPN that does not apply to any order number - invoice "%s", txn_type "%s"' % (this_notification.invoice, this_notification.txn_type))
        return 
        
    def compare_cart(self, notification, order):
        for ref in order.shopping_cart.keys():
            cart_item = order.shopping_cart[ref]
            if notification.shopping_cart.has_key(cart_item.product_code):
                notification_item = notification.shopping_cart[cart_item.product_code]
                if int(cart_item.quantity) != int(notification_item.quantity):
                    return False
            else:
                # item not in returned cart - invalid IPN response
                return False
        # everything checks out
        return True
            


    def verify(self):
        options = IPaypalStandardOptions( self.portal )
        # get ready to POST back form variables
        
        form = self.request.form
        params =[(key, form[key]) for key in form.keys() if key != 'cmd']
        params = [('cmd', '_notify-validate')] + params
        paramData = urllib.urlencode(params)
        url = "https://%s/cgi-bin/webscr" % _sites[options.server_url]
        
        # TODO: is this dangerous for other products?
        socket.setdefaulttimeout(120)
        try:
            connection = urllib2.urlopen(url, paramData)
        except urllib2.error.URLError:
            raise IOError, 'IPN post-back failed'
        response = connection.read()
        connection.close()
               
        if response == 'VERIFIED':
            return True
        else:
            return False

    def fill_in_order_data(self, notification, order):

        name = ''
        if notification.last_name is not None and notification.first_name is not None:
            name = "%s %s" % (notification.first_name, notification.last_name)
        else:
            if notification.first_name is not None:
                name = notification.last_name
            else:
                name = notification.first_name
        order.contact_information.name = name
        
        if notification.email is not None:
            order.contact_information.email = notification.email
        
        if notification.phone is not None:
            order.contact_information.phone_number = re.sub('[^\d]+', '', notification.phone)
        
        if notification.address_street is not None:
            order.billing_address.bill_first_line = notification.address_street
        
        if notification.address_city is not None:
            order.billing_address.bill_city = notification.address_city
        
        if notification.address_state is not None:
            order.billing_address.bill_state = notification.address_state

        if notification.address_zip is not None:
            order.billing_address.bill_postal_code = notification.address_zip

        if notification.address_country is not None:
            order.billing_address.bill_country = notification.address_country

        # mark address as same - Paypal does not provide seperate shipping address
        order.shipping_address.ship_same_billing = True


"""
"""

from getpaid.core.options import PersistentOptions
import interfaces

PaypalStandardOptions = PersistentOptions.wire("PaypalStandardOptions",
                                               "getpaid.paypal",
                                               interfaces.IPaypalStandardOptions )




from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope.component import getUtility

from Products.PloneGetPaid.interfaces import IGetPaidManagementOptions
from getpaid.core.interfaces import IShoppingCartUtility, IOrderManager
from getpaid.core.order import Order
from getpaid.core import payment

from cPickle import loads, dumps
from AccessControl import getSecurityManager

from getpaid.paypal.paypal import PaypalStandardProcessor

class PaypalButtonView(BrowserView):
    """page for paypal button
    """
    
    def __init__(self, context, request):
        self.context = context
        self.request = request
    
    def getButton(self):
        button = PaypalStandardProcessor(self.context)
        cart_util = getUtility(IShoppingCartUtility)
        cart = cart_util.get(self.context, create=True)
        manage_options = IGetPaidManagementOptions( self.context )
        
        # we'll get the order_manager, create the new order, and store it.
        order_manager = getUtility(IOrderManager)
        new_order_id = order_manager.newOrderId()
        order = Order()
        
        order.finance_workflow.fireTransition('create')
        
        # register the payment processor name to make the workflow handlers happy
        order.processor_id = manage_options.payment_processor
        
        # FIXME: registering an empty contact information list for now - need to populate this from user
        # if possible
        order.contact_information = payment.ContactInformation()
        order.billing_address = payment.BillingAddress()
        order.shipping_address = payment.ShippingAddress()

        order.order_id = new_order_id
        
        # make cart safe for persistence by using pickling
        order.shopping_cart = loads(dumps(cart))
        order.user_id = getSecurityManager().getUser().getId()
        order_manager.store(order)

        # have to wait for the order to be created and the cart added for this to work
        order.finance_workflow.fireTransition('authorize')

        # save html for button - we'll destroy the cart later on
        html = button.cart_post_button(order)
        
        # and destroy the cart
        cart_util.destroy(self.context)

        return html
        


from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.CMFCore.utils import getToolByName

class GetpaidPaypalThankyouView(BrowserView):
    """Class for overriding getpaid-thank-you view for paypal purchases
    """
    
    def __init__(self, context, request):
        self.context = context
        self.request = request

    def getInvoice(self):
        if self.request.has_key('invoice'):
            return self.request['invoice']
        else:
            return None

    def getURL(self):
        portalurl = getToolByName(self.context, "portal_url").getPortalObject().absolute_url()
        if self.getInvoice() is not None:
            return "%s/@@getpaid-order/%s" % ( portalurl, self.getInvoice())
        else:
            return ''